﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
PowerHandles v0.2 by MacGruber.
Improved Position/Rotation handles for VaM.

Version 0.2 2021-07-17
	Fixed handles not working when "Only Show Targets and Overlays When Menu Open" is disabled in UserPreferences and the UI is closed.
	
Version 0.1 2021-01-17
	Initial release.
	
///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using Battlehub.RTCommon;
using GridMode = FreeControllerV3.GridMode;
using System.Collections.Generic;

using System.Text;
using UnityEngine.UI;
using UnityEngine.Events;

namespace MacGruber
{
    public class PowerHandles : MVRScript
    {
		private Transform myMouseGrab;
		private FreeControllerV3 myController;
		private GridMode myPositionGridMode = GridMode.None;
		private GridMode myRotationGridMode = GridMode.None;
		private float myPositionGridAlign;
		private float myRotationGridAlign;
		
		private JSONStorableFloat myPositionGridAlignUI;
		private JSONStorableFloat myRotationGridAlignUI;
		private JSONStorableFloat myGizmoScaleUI;
		
		public override void Init()
		{
			SetupInfoText(
				"<color=#606060><size=40><b>PowerHandles v0.2</b></size>\nImproved Position/Rotation handles for VaM Desktop Mode.</color>\n\n" +
				"<b>Usage:</b>\nLeftCTRL:  Enable Position\nLeftALT:     Enable Rotation\nLeftSHIFT: Global Coordinates\nSPACE:     Enable Grid Mode",
				350.0f, true
			);
			
			myGizmoScaleUI = SetupSliderFloat("Gizmo Scale", 1.0f, 0.05f, 1.0f, false);
			myGizmoScaleUI.setCallbackFunction += SyncMeshScale;
			SyncMeshScale(myGizmoScaleUI.val);
						
			myPositionGridAlignUI = SetupSliderFloat("Position Grid", 0.1f, 0.01f, 1.0f, false);
			myPositionGridAlignUI.setCallbackFunction += (float v) => {
				v = Mathf.Round(v*100.0f)/100.0f;
				myPositionGridAlignUI.valNoCallback = v;				
				if (myController != null)
					myController.positionGrid = v;
			};			
			
			myRotationGridAlignUI = SetupSliderInt("Rotation Grid", 15, 1, 90, false);
			myRotationGridAlignUI.setCallbackFunction += (float v) => {
				if (myController != null)
					myController.rotationGrid = v;
			};
		}
		
		private void Update()
		{	
			SuperController sc = SuperController.singleton;
			UserPreferences up = UserPreferences.singleton;
			bool handlesEnabled = sc.MonitorCenterCamera != null && (!up.showTargetsMenuOnly || sc.mainHUD.gameObject.activeSelf);
			bool usePosition = handlesEnabled && Input.GetKey(KeyCode.LeftControl);
			bool useRotation = handlesEnabled && Input.GetKey(KeyCode.LeftAlt);
			bool useGlobal = handlesEnabled && Input.GetKey(KeyCode.LeftShift);
			bool useGrid = handlesEnabled && Input.GetKey(KeyCode.Space);
			
			FCPositionHandle positionHandle = sc.selectedControllerPositionHandle;
			FCRotationHandle rotationHandle = sc.selectedControllerRotationHandle;
			if (positionHandle != null)
			{
				if (positionHandle.enabled && !usePosition)
					positionHandle.ForceDeselectAxis();
				positionHandle.enabled = usePosition;
			}
			if (rotationHandle != null)
			{
				if (rotationHandle.enabled && !useRotation)
					rotationHandle.ForceDeselectAxis();
				rotationHandle.enabled = useRotation;
			}
			
			RuntimeTools.PivotRotation = useGlobal ? RuntimePivotRotation.Global : RuntimePivotRotation.Local;
			
			FreeControllerV3 controller = sc.GetSelectedController();
			if (controller != myController)
			{
				SelectController(controller);
			}
			if (myController != null)
			{
				GridMode gridMode = GridMode.None;				
				if (useGrid)
					gridMode = useGlobal ? GridMode.Global : GridMode.Local;
				if (myController.positionGridMode != gridMode)
					myController.positionGridMode = gridMode;			
				if (myController.rotationGridMode != gridMode)
					myController.rotationGridMode = gridMode;
			}
		}
		
		private void OnAtomAdded(Atom atom)
		{
			if (myGizmoScaleUI != null)
				SyncMeshScale(myGizmoScaleUI.val);
		}
		
		private void OnEnable()
		{
			SuperController sc = SuperController.singleton;	
			
			myMouseGrab = sc.mouseGrab;
			sc.mouseGrab = null;
			
			sc.onAtomAddedHandlers += OnAtomAdded;
			
			if (myGizmoScaleUI != null)
				SyncMeshScale(myGizmoScaleUI.val);
		}
		
		private void OnDisable()
		{
			ResetController();
			
			SuperController sc = SuperController.singleton;
			bool handlesEnabled = sc.mainHUD.gameObject.activeSelf && sc.MonitorCenterCamera != null;
			
			FCPositionHandle positionHandle = sc.selectedControllerPositionHandle;
			FCRotationHandle rotationHandle = sc.selectedControllerRotationHandle;
			if (positionHandle != null)
			{
				positionHandle.ForceDeselectAxis();
				positionHandle.enabled = handlesEnabled;
			}
			if (rotationHandle != null)
			{
				rotationHandle.ForceDeselectAxis();
				rotationHandle.enabled = handlesEnabled;
			}
			
			RuntimeTools.PivotRotation = RuntimePivotRotation.Local;
				
			sc.mouseGrab = myMouseGrab;
			myMouseGrab = null;
			
			sc.onAtomAddedHandlers -= OnAtomAdded;
			SyncMeshScale(1.0f);
		}
		
		private void SelectController(FreeControllerV3 controller)
		{
			ResetController();
			myController = controller;
			if (myController != null)
			{
				myPositionGridMode = myController.positionGridMode;
				myRotationGridMode = myController.rotationGridMode;
				myPositionGridAlign = myController.positionGrid;
				myRotationGridAlign = myController.rotationGrid;
				myController.positionGrid = myPositionGridAlignUI.val;
				myController.rotationGrid = myRotationGridAlignUI.val;
			
				// Hack to hide collider mesh of Empty atoms when collision is disabled
				if (myController.containingAtom.type == "Empty" && myController.containingAtom.collisionEnabledJSON != null)
				{
					if (!myController.containingAtom.collisionEnabled)
						SyncCollisionEnabled(false);
					myController.containingAtom.collisionEnabledJSON.setCallbackFunction += SyncCollisionEnabled;
				}
			}
		}
		
		private void ResetController()
		{
			if (myController != null)
			{
				myController.positionGridMode = myPositionGridMode;
				myController.rotationGridMode = myRotationGridMode;
				myController.positionGrid = myPositionGridAlign;
				myController.rotationGrid = myRotationGridAlign;
				
				if (myController.containingAtom.type == "Empty" && myController.containingAtom.collisionEnabledJSON != null)
				{
					myController.containingAtom.collisionEnabledJSON.setCallbackFunction -= SyncCollisionEnabled;
				}
				
				myController = null;	
			}
		}
		
		private void SyncCollisionEnabled(bool v)
		{
			Transform t = myController?.containingAtom?.reParentObject?.Find("object/rescaleObject/SphereRender");
			if (t != null)
				t.gameObject.SetActive(v);
		}
		
		private void SyncMeshScale(float scale)
		{
			List<FreeControllerV3> controllers = SuperController.singleton.GetAllFreeControllers();
			for (int i=0; i<controllers.Count; ++i)
			{
				FreeControllerV3 c = controllers[i];
				c.deselectedMeshScale = 0.02f * scale;
				c.meshScale = 0.60f * scale;
			}
		}
				
		private JSONStorableFloat SetupSliderFloat(string name, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(name, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			CreateSlider(storable, rightSide);
			RegisterFloat(storable);
			return storable;
		}
		
		private JSONStorableFloat SetupSliderInt(string label, int defaultValue, int minValue, int maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(label, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			UIDynamicSlider slider = CreateSlider(storable, rightSide);
			slider.slider.wholeNumbers = true;
			slider.valueFormat = "F0";
			RegisterFloat(storable);
			return storable;
		}
				
		private JSONStorableString SetupInfoText(string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = CreateTextField(storable, rightSide);
			textfield.height = height;
			return storable;
		}
    }
}
