﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
SkyMagic Exporter v0.4 by MacGruber.
Takes a cube-screenshot and saves it in SMC format. You will need to import this into Unity using SkyMagicImporter, then export back using an Unity AssetBundle, so you can load it again in VaM using SkyMagic Loader.

Version 0.4 2021-01-30
	Now saving cubeshots by to 'Saves/PluginData/SkyMagic' instead of 'Saves/screenshots' to comply with VaM's new plugin guideline.

Version 0.3 2020-10-10
	No changes except matching version to SkyMagicLoader.

Version 0.2 2020-05-01
	Direct PNG export.
	Fixing rendering of characters and other native VaM objects.
	Support for PostMagic.

Version 0.1 2020-04-12
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.Rendering;
using UnityEngine.PostProcessing;
using System.Collections;
using mset;
using MVR.FileManagementSecure;
using System;
using Battlehub.RTCommon;

namespace MacGruber
{
    public class SkyMagicExporter : MVRScript
    {	
		private static readonly string SAVE_DIRECTORY = "Saves/PluginData/SkyMagic/";
	
		private JSONStorableFloat myResolution;
		
		
		public override void Init()
		{		
			SetupInfoText(this,
				"<color=#606060><size=40><b>SkyMagic Exporter v0.4</b></size>\nTakes a cube-screenshot and saves it in SMC.PNG format into this folder:\n<i>Saves/PluginData/SkyMagic/</i>\n\nYou will need to import this into Unity using SkyMagicImporter, then export back using an Unity AssetBundle, so you can load it again in VaM using SkyMagic Loader.</color>\n\n",
				600.0f, true
			);
			
			myResolution = SetupSliderFloat(this, "Resolution", 256, 64, 1024, false);
			myResolution.setCallbackFunction  += (float v) => { 
				myResolution.valNoCallback = Mathf.ClosestPowerOfTwo(Mathf.RoundToInt(v));
			};
				
			SetupButton(this, "Export Cubemap", FileBrowser, false);
		}
			
		private void OnDestroy()
		{
			GameObject cubemapGameObject = GameObject.Find("SkyMagicCamera");			
			if (cubemapGameObject != null)
				Destroy(cubemapGameObject);
		}
			
		private void FileBrowser()
		{
			FileManagerSecure.CreateDirectory(SAVE_DIRECTORY);	
		
			SuperController sc = SuperController.singleton;			
			sc.GetMediaPathDialog(ExportCubemap, "png", SAVE_DIRECTORY, false, true, false, null, false, null, false, false);
			sc.mediaFileBrowserUI.SetTextEntry(true);
			if (sc.mediaFileBrowserUI.fileEntryField != null)
			{
				string filename = ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString();
				sc.mediaFileBrowserUI.fileEntryField.text = filename + ".smc.png";
				sc.mediaFileBrowserUI.ActivateFileNameField();
			}
		}
			
		private void ExportCubemap(string path)
		{
			StartCoroutine(ExportCubemapCO(path));
		}
		
		private IEnumerator ExportCubemapCO(string path)
		{
			if (!path.EndsWith(".png"))
				path += ".smc.png";
			else if (!path.EndsWith(".smc.png"))
				path = path.Substring(0, path.Length-4) + ".smc.png";
			
			SuperController sc = SuperController.singleton;
			sc.ClearSelection(); // clear selection to avoid UI elements in the camera image.
			
			// get a camera
			Camera screenshotCamera = sc.hiResScreenshotCamera;
			GameObject cubemapGameObject = GameObject.Find("SkyMagicCamera");
			Camera[] cubemapCameras;
			if (cubemapGameObject == null)
			{
				cubemapGameObject = new GameObject("SkyMagicCamera");
				cubemapCameras = new Camera[6];
				for (int i=0; i<6; ++i)
				{
					Camera c = Instantiate<Camera>(screenshotCamera, cubemapGameObject.transform);
					MoveAndRotateAs mr = c.GetComponent<MoveAndRotateAs>();
					if (mr)
						mr.MoveAndRotateAsTransform = null;
					c.gameObject.name = "Camera"+(i+1);
					c.enabled = true;					
					cubemapCameras[i] = c;
				}
			}
			else
			{
				cubemapCameras = cubemapGameObject.GetComponentsInChildren<Camera>();
			}			
			
			// setup cameras
			int cullingMask = 1 << LayerMask.NameToLayer("UI")
							| 1 << LayerMask.NameToLayer("LoadUI")
							| 1 << LayerMask.NameToLayer("ScreenUI")
							| 1 << LayerMask.NameToLayer("GUI");
			cullingMask = screenshotCamera.cullingMask & ~cullingMask;
			int s = Mathf.ClosestPowerOfTwo(Mathf.RoundToInt(myResolution.val));
			RenderTexture rt = RenderTexture.GetTemporary(s*6, s, 24, RenderTextureFormat.ARGB32, RenderTextureReadWrite.sRGB, 1);
			
			Quaternion[] rotations = new Quaternion[6] {
				Quaternion.Euler(0, 90,0),
				Quaternion.Euler(0,-90,0),
				Quaternion.Euler(-90,0,0),
				Quaternion.Euler( 90,0,0),
				Quaternion.Euler(0,  0,0),
				Quaternion.Euler(0,180,0)
			};
			for (int i=0; i<6; ++i)
			{
				Camera c = cubemapCameras[i];
				c.transform.localPosition = Vector3.zero;
				c.transform.localRotation = rotations[i];
				c.fieldOfView = 90;
				c.cullingMask = cullingMask;
				c.targetTexture = rt;
				c.pixelRect = new Rect(i*s, 0, s, s);
				
				// compatibility with PostMagic
				PostProcessingBehaviour screenshotBehaviour = screenshotCamera.GetComponent<PostProcessingBehaviour>();
				PostProcessingBehaviour cubemapBehaviour = c.GetComponent<PostProcessingBehaviour>();
				if (screenshotBehaviour == null)
				{
					if (cubemapBehaviour != null)
						cubemapBehaviour.enabled = false;
				}
				else
				{
					if (cubemapBehaviour == null)
						cubemapBehaviour = c.gameObject.AddComponent<PostProcessingBehaviour>();
					cubemapBehaviour.enabled = true;
					cubemapBehaviour.profile = screenshotBehaviour.profile;
				}
			}
			cubemapGameObject.transform.position = containingAtom.mainController.transform.position;
			cubemapGameObject.transform.rotation = Quaternion.identity;
			
			// wait for cube-screenshot
			cubemapGameObject.SetActive(true);
			yield return null;			
			cubemapGameObject.SetActive(false);
			
			// save image
			Texture2D texture = new Texture2D(s*6, s, TextureFormat.RGB24, false);			
			RenderTexture previousRT = RenderTexture.active;
			RenderTexture.active = rt;
			Rect sourceRect = new Rect(0, 0, s*6, s);
			texture.ReadPixels(sourceRect, 0, 0, false);
			texture.Apply();
			RenderTexture.active = previousRT;			
			byte[] bytes = texture.EncodeToPNG();
			FileManagerSecure.WriteAllBytes(path, bytes);
							
			// cleanup
			RenderTexture.ReleaseTemporary(rt);
			Destroy(texture);
			
			sc.SelectController(containingAtom.mainController);
		}
						
		private static void SetupButton(MVRScript script, string label, UnityAction callback, bool rightSide)
		{
			UIDynamicButton button = script.CreateButton(label, rightSide);
			button.button.onClick.AddListener(callback);
		}
		
		private static JSONStorableFloat SetupSliderFloat(MVRScript script, string label, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(label, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateSlider(storable, rightSide);
			script.RegisterFloat(storable);
			return storable;
		}
		
		private static JSONStorableString SetupInfoText(MVRScript script, string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = script.CreateTextField(storable, rightSide);
			textfield.height = height;
			return storable;
		}
    }
}
