﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
SuperResolution v12 by MacGruber.
Makes the VaM internal screenshot tool do screenshots in 8K resolution.

Version 12 2021-06-12
	Changes to adresse issues introduced by VaM 1.20.77.2 update:
		SuperVideo: Fixed AspectRatio (of your VaM window) not being applied to preview display and HUD preview.
		SuperVideo: Fixed "Preview UI Scale" slider not being applied to preview display.
		SuperVideo: Removed "Preview UI Left/Right" slider as it was obsolete, since VaM 1.20.77.2 you can just move the DisplayControl around.
	Updating to MacGruber_Utils.cs 2021-05-15.

Version 11 2021-01-30
	SuperShot: Minor UI cleanup/improvement.

Version 10 2021-01-16
	SuperShot: Adding 32:9 and 36:9 resolutions.

Version 0.9 2020-10-10
	Display log message if the plugin or atom is not enabled to avoid confusion of silently doing nothing.

Version 0.8 2020-08-14
	Fixed not saving screenshots when directory doesn't exist.

Version 0.7 2020-08-13
	Fixing interaction with PostMagic. Note that screenspace post-effects can behave differently when applied at higher resolution.
	SuperShot: Clear selection while taking screenshot to avoid UI elements in the camera image.	
	SuperShot: Fixed memory leak when not using super-sampling.
	SuperShot: Custom camera feature. Take multiple screenshots from different perspectives at once.
	SuperShot: Screenshots can be triggered by VaM action. You can hook it to your custom buttons or hotkeys.
	SuperVideo: Removed additional blit step for better performance.

Version 0.6 2020-08-05
	Implementing hook into WindowCamera (=> SuperVideo)
	Implementing hook into Screenshot camera (=> SuperShot).
	Implementing aspect ratio choice.
	Implementing downscale shader.
	Implementing PNG/JPG choice.
	Implementing screenshot from player perspective.
	Implementing improved preview UI (scale, offset, etc).
	Implementing Lock Z Rotation
	Implementing optional flash light.
	Implementing memory estimate.
	Massive cleanup/refactor.

Version 0.5 2020-03-21
	Adding to MacGruber-Essentials Package.

Version 0.4 2019-10-12
	Fixing FoV issue with the preview camera.
	Compatibility with PostMagic.
	Added override for MSAA setting applied to screenshots.

Version 0.3 2019-10-04
	Fixed VR mode screenshots issue with VaM 1.18

Version 0.2 2019-08-18
	Adding preview camera to allow use in VR mode without performance issues.
	Adding slider with resolutions to select from.

Version 0.1 2019-08-16
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */


using UnityEngine;
using System.Collections.Generic;
using System;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;

namespace MacGruber
{
	public class SuperResolution
	{
		public static readonly string SCREENSHOT_DIRECTORY   = "Saves/screenshots/";
		
		public const int MAX_RESOLUTION = 8192;

		public static readonly List<int> MULTIPLIER_VALUES = new List<int>() { 1, 2, 3, 4 };
		public static readonly List<string> MULTIPLIER_NAMES = new List<string>() { "Off", "2x", "3x", "4x" };

		public static readonly List<string> KERNEL_NAMES = new List<string>() { "Linear (Blurring)", "Lanczos 3 (Sharpening)" };
		public const int KERNEL_LINEAR = 0;
		public const int KERNEL_LANCZOS3 = 1;
		public const int KERNEL_LANCZOS2 = 2;

		public static readonly List<int> MSAA_VALUES = new List<int>() { 1, 2, 4, 8 };
		public static readonly List<string> MSAA_NAMES = new List<string>() { "Off", "2x", "4x", "8x" };

		public static readonly List<string> FORMAT_NAMES = new List<string>() { "PNG (Quality)", "JPG (Smaller)" };
		public const int FORMAT_PNG = 0;
		public const int FORMAT_JPG = 1;

		public int myOutputWidth = 1920;
		public int myOutputHeight = 1080;
		public int myMultiplier = 4;
		public int myMsaaLevel = 8;
		public int myKernelMode = KERNEL_LANCZOS3;
		public bool myOutputToScreen = false;

		private bool myNeedInit = true;
		private bool myNeedSetup = true;

		public RenderTexture GetCaptureTexture()
		{
			return myCaptureTexture;
		}

		public RenderTexture GetOutputTexture()
		{
			return myOutputTexture;
		}

		private RenderTexture myCaptureTexture;
		private RenderTexture myPassTexture;
		private RenderTexture myOutputTexture;
		private Material myConvolutionMaterial;
		private string myShaderBundleURL = null;



		public void OnInit(MVRScript script)
		{
			myNeedInit = true;
			myNeedSetup = true;
			myShaderBundleURL = Utils.GetPluginPath(script) + "/MacGruber_Convolution.shaderbundle";

			Request request = new AssetLoader.AssetBundleFromFileRequest {path = myShaderBundleURL, callback = OnAssetBundleLoaded};
			AssetLoader.QueueLoadAssetBundleFromFile(request);
		}

		private void OnAssetBundleLoaded(Request aRequest)
		{
			myConvolutionMaterial = aRequest.assetBundle?.LoadAsset<Material>("Assets/SuperResolution/Convolution.mat");

			if (myConvolutionMaterial == null)
				SuperController.LogError("SuperResolution: Could not load convolution shader!");

			myNeedInit = myConvolutionMaterial == null;
		}

		public void OnDestroy()
		{
			if (myShaderBundleURL != null)
			{
                AssetLoader.DoneWithAssetBundleFromFile(myShaderBundleURL);
				myShaderBundleURL = null;
			}
			myNeedInit = true;
		}

		public void OnDisable()
		{
			myNeedSetup = true;

			if (myCaptureTexture != myOutputTexture)
				ReleaseRenderTexture(myCaptureTexture);
			ReleaseRenderTexture(myPassTexture);
			ReleaseRenderTexture(myOutputTexture);
			myCaptureTexture = null;
			myPassTexture = null;
			myOutputTexture = null;
		}

		public void NeedSetup()
		{
			myNeedSetup = true;
		}

		private void Setup()
		{
			if (myNeedInit)
				return;

			// setup material coefficients
			const int maxCoefficients = 24;
			float[] coefficients = new float[maxCoefficients];
			if (myKernelMode == KERNEL_LINEAR)
			{
				for (int i=0; i<myMultiplier; ++i)
					coefficients[i] = 1.0f / myMultiplier;
				for (int i=myMultiplier; i<maxCoefficients; ++i)
					coefficients[i] = 0.0f;
				myConvolutionMaterial.SetFloatArray("_Coefficients", coefficients);
				myConvolutionMaterial.SetInt("_CoefficientsCount", myMultiplier);
			}
			else // KERNEL_LANCZOS2 or KERNEL_LANCZOS3
			{
				float a = (myKernelMode == KERNEL_LANCZOS2) ? 2 : 3;
				float kernelWidth = 4 * myMultiplier;
				int kernelSamples = (int)Mathf.Ceil(2.5f*myMultiplier);
				float kernelScale = a / (kernelWidth*0.5f);

				float sum = 0;
				for (int i=0; i<kernelSamples; ++i)
				{
					float x = i - kernelSamples*0.5f;
					float x1 = kernelScale * x;
					float x2 = kernelScale * (x + 1.0f);
					float k = LanczosIntegral(a, x1, x2);
					coefficients[i] = k;
					sum += k;
				}
				for (int i=0; i<kernelSamples; ++i)
					coefficients[i] /= sum;
				for (int i=kernelSamples; i<maxCoefficients; ++i)
					coefficients[i] = 0.0f;
				myConvolutionMaterial.SetFloatArray("_Coefficients", coefficients);
				myConvolutionMaterial.SetInt("_CoefficientsCount", kernelSamples);
			}

			myNeedSetup = false;
		}

		public void PreRender()
		{
			if (myNeedSetup)
				Setup();
					
			if (myOutputToScreen && myMultiplier == 1)
				myCaptureTexture = null;
			else				
				myCaptureTexture = GetRenderTexture(myOutputWidth*myMultiplier, myOutputHeight*myMultiplier, 24, myMsaaLevel);
			
			myOutputTexture = myMultiplier == 1 ? myCaptureTexture : null;
		}

		public void PostRender()
		{
			if (myCaptureTexture != myOutputTexture)
				ReleaseRenderTexture(myOutputTexture);
			ReleaseRenderTexture(myCaptureTexture);
			myCaptureTexture = null;
			myOutputTexture = null;
		}

		public void ProcessImage()
		{
			if (myNeedInit)
				return;

			if (myMultiplier > 1)
			{
				// downscale
				int m = myMultiplier;
				Vector4 scaleWidth = new Vector4(1.0f/myCaptureTexture.width,0,0,0);
				Vector4 scaleHeight = new Vector4(0,1.0f/myCaptureTexture.height,0,0);
				bool horizontal = myCaptureTexture.width >= myCaptureTexture.height;

				myConvolutionMaterial.SetVector("_TexelSize", horizontal ? scaleWidth : scaleHeight);
				myPassTexture = GetRenderTexture(myOutputWidth*(horizontal?1:m), myOutputHeight*(horizontal?m:1));
				Graphics.Blit(myCaptureTexture, myPassTexture, myConvolutionMaterial, 0);				
				ReleaseRenderTexture(myCaptureTexture);
				myCaptureTexture = null;

				myConvolutionMaterial.SetVector("_TexelSize", horizontal ? scaleHeight : scaleWidth);
				myOutputTexture = myOutputToScreen ? null : GetRenderTexture(myOutputWidth, myOutputHeight);
				Graphics.Blit(myPassTexture, myOutputTexture, myConvolutionMaterial, 0);
				ReleaseRenderTexture(myPassTexture);
				myPassTexture = null;
			}
			else
			{
				myOutputTexture = myOutputToScreen ? null : myCaptureTexture;
			}
		}

		public void SaveImage(int fileFormat = FORMAT_PNG, string overrideFilename = null)
		{
			if (myNeedInit)
				return;

			// save image
			Texture2D texture = new Texture2D(myOutputTexture.width, myOutputTexture.height, TextureFormat.RGB24, false);
			RenderTexture previousRT = RenderTexture.active;
			RenderTexture.active = myOutputTexture;
			Rect sourceRect = new Rect(0, 0, myOutputTexture.width, myOutputTexture.height);
			texture.ReadPixels(sourceRect, 0, 0, false);
			texture.Apply(false);
			RenderTexture.active = previousRT;
			
			string filename = overrideFilename == null ? GetFilename() : overrideFilename;
			filename += (fileFormat == FORMAT_JPG) ? ".jpg" : ".png";
			byte[] bytes = (fileFormat == FORMAT_JPG) ? texture.EncodeToJPG(100) : texture.EncodeToPNG();
			
			FileManagerSecure.CreateDirectory(SCREENSHOT_DIRECTORY);				
			FileManagerSecure.WriteAllBytes(SCREENSHOT_DIRECTORY + filename, bytes);
		}
		
		public string GetFilename()
		{
			return ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString();
		}
		
		public float EstimatedMemoryUseInGB()
		{
			long size = 0;
			// capture texture
			size = (long)(myOutputWidth * myOutputHeight) * (long)(myMultiplier * myMultiplier * myMsaaLevel * 8);
			if (myMsaaLevel > 1)
			{
				// MSAA resolve
				size += (long)(myOutputWidth * myOutputHeight) * (long)(myMultiplier * myMultiplier * 8);
			}
			if (myMultiplier > 1)
			{
				// downscale
				size += (long)(myOutputWidth * myOutputHeight) * (long)(myMultiplier * 4);
				size += (long)(myOutputWidth * myOutputHeight) * 4L;
			}

			size /= 1024*1024;
			float sizeInMB = size * 1.333f;  // somehow we seem to use 1/3 more...maybe mipmaps, although they are disabled?
			return Mathf.Max(Mathf.Round(sizeInMB / 102.4f) / 10.0f, 0.1f);
		}

		public RenderTexture GetPreviewTexture(int width, int height)
		{
			RenderTextureDescriptor descriptor = new RenderTextureDescriptor(width, height, RenderTextureFormat.ARGB32, 24);
			descriptor.msaaSamples = 1;
			descriptor.autoGenerateMips = true;
			descriptor.useMipMap = true;
			descriptor.sRGB = true;
			RenderTexture texture = RenderTexture.GetTemporary(descriptor);
			texture.filterMode = FilterMode.Trilinear;
			texture.wrapMode = TextureWrapMode.Clamp;
			return texture;
		}

		public RenderTexture GetRenderTexture(int width, int height, int depthBufferBits = 0, int msaa = 1)
		{
			RenderTextureDescriptor descriptor = new RenderTextureDescriptor(width, height, RenderTextureFormat.ARGB32, depthBufferBits);
			descriptor.msaaSamples = msaa;
			descriptor.autoGenerateMips = false;
			descriptor.useMipMap = false;
			descriptor.sRGB = true;
			RenderTexture texture = RenderTexture.GetTemporary(descriptor);
			texture.filterMode = FilterMode.Point;
			texture.wrapMode = TextureWrapMode.Clamp;
			return texture;
		}

		public void ReleaseRenderTexture(RenderTexture texture)
		{
			if (texture != null)
				RenderTexture.ReleaseTemporary(texture);
		}

		private static float Lanczos(float a, float x)
		{
			if (-0.001f < x && x < 0.001f)
				return 1.0f;
			else if (x <= -a || x >= a)
				return 0.0f;

			float xPi = Mathf.PI * x;
			return a * Mathf.Sin(xPi) * Mathf.Sin(xPi/a) / (xPi*xPi);
		}

		private static float LanczosIntegral(float a, float x1, float x2)
		{
			const int integralSamples = 8;
			float sum = 0;
			for (int i=0; i<=integralSamples; ++i)
			{
				float x = Mathf.Lerp(x1, x2, (float)i/integralSamples);
				sum += Lanczos(a, x);
			}
			return sum / integralSamples;
		}
	}
}