﻿using UnityEngine;
using UnityEngine.XR;
using UnityEngine.UI;
using UnityEngine.Events;
using UnityEngine.PostProcessing;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System;
using MeshVR;
using static MacGruber.SuperResolution;


namespace MacGruber
{
	public class SuperShot : MVRScript
	{
		private const int MODE_NONE = 0;
		private const int MODE_BUILDIN = 1;
		private const int MODE_PLAYER = 2;
		private const int STATE_SETUP = 0;
		private const int STATE_WAIT = 1;
		private const int STATE_PROCESS = 2;

		private static readonly List<string> ASPECTRATIO_NAMES = new List<string>() {
			"36:9",
			"32:9",
			"21:9",
			"16:9",
			"16:10",
			"4:3",
			"1:1"
		};

		private static readonly List<Vector2Int>[] RESOLUTION_VALUES = new List<Vector2Int>[] {
			new List<Vector2Int>() { // 36:9
				new Vector2Int(1600, 400),
				new Vector2Int(3200, 800),
				new Vector2Int(6400, 1600)
			},
			new List<Vector2Int>() { // 32:9
				new Vector2Int(2048, 576),
				new Vector2Int(2560, 720),
				new Vector2Int(3840, 1080),
				new Vector2Int(5120, 1440),
				new Vector2Int(7680, 2160),
			},
			new List<Vector2Int>() { // 21:9
				new Vector2Int(2560, 1080),
				new Vector2Int(3440, 1440),
				new Vector2Int(5120, 2160)
			},
			new List<Vector2Int>() { // 16:9
				new Vector2Int(1280, 720),
				new Vector2Int(1920, 1080),
				new Vector2Int(2560, 1440),
				new Vector2Int(3840, 2160),
				new Vector2Int(7680, 4320)
			},
			new List<Vector2Int>() { // 16:10
				new Vector2Int(1280, 800),
				new Vector2Int(1440, 900),
				new Vector2Int(1920, 1200),
				new Vector2Int(3840, 2400),
				new Vector2Int(7680, 4800)
			},
			new List<Vector2Int>() { // 4:3
				new Vector2Int(800, 600),
				new Vector2Int(1024, 768),
				new Vector2Int(2048, 1536),
				new Vector2Int(4096, 3072),
				new Vector2Int(8192, 6144),
			},
			new List<Vector2Int>() { // 1:1
				new Vector2Int(256, 256),
				new Vector2Int(512, 512),
				new Vector2Int(1024, 1024),
				new Vector2Int(2048, 2048),
				new Vector2Int(4096, 4096),
				new Vector2Int(8192, 8192)
			}
		};

		private static readonly List<string>[] RESOLUTION_NAMES = new List<string>[] {
			new List<string>() { // 36:9
				"1600x400 (Patreon Cover)",
				"3200x800 (HQ Patreon Cover)",
				"6400x1600"
			},
			new List<string>() { // 32:9
				"2048x576",
				"2560x720",
				"3840x1080 (DFHD)",
				"5120x1440 (DQHD)",
				"7680x2160 (DUHD)",
			},
			new List<string>() { // 21:9
				"2560x1080 (WFHD)",
				"3440x1440 (WQHD)",
				"5120x2160 (4K WUHD)"
			},
			new List<string>() { // 16:9
				"1280x720 (HD)",
				"1920x1080 (FHD)",
				"2560x1440 (QHD)",
				"3840x2160 (4K UHD)",
				"7680x4320 (8K UHD)"
			},
			new List<string>() { // 16:10
				"1280x800 (WXGA)",
				"1440x900 (WXGA+)",
				"1920x1200 (WUXGA)",
				"3840x2400 (2x WUXGA)",
				"7680x4800 (4x WUXGA)"
			},
			new List<string>() { // 4:3
				"800x600 (SVGA)",
				"1024x768 (XGA)",
				"2048x1536 (2x XGA)",
				"4096x3072 (4x XGA)",
				"8192x6144 (8x XGA)"
			},
			new List<string>() { // 1:1
				"256x256",
				"512x512",
				"1024x1024",
				"2048x2048",
				"4096x4096",
				"8192x8192"
			}
		};

		private static readonly int[] DEFAULT_RESOLUTION_IDX = new int[] { 1, 1, 0, 1, 2, 1, 2 };

		private const int DEFAULT_ASPECTRATIO_IDX = 3;
		private const int DEFAULT_MULTIPLIER_IDX = 3;
		private const int DEFAULT_KERNEL = KERNEL_LANCZOS3;
		private const int DEFAULT_MSAA_IDX = 3;
		private const int DEFAULT_FORMAT = FORMAT_PNG;

		private JSONStorableBool myPortaitMode;
		private JSONStorableFloat myPreviewScale;
		private JSONStorableBool myLockZRotation;
		private JSONStorableBool myFlashLight;
		private JSONStorableString myMemoryInfo;
		private float myMemoryEstimate = -1.0f;

		private Camera myPreviewCamera;
		private Camera myScreenshotCamera;
		private ScreenshotCameraHook myScreenshotHook;
		private PreviewCameraHook myPreviewHook;
		private RenderTexture myOriginalScreenshotTexture;
		
		private Transform myOverrideTransform;
		private float myOverrideFoV = -1.0f;
		private string myOverrideFilename = null;

		private int myMode = MODE_NONE;
		private int myState = STATE_SETUP;
		private int myAspectRatioIdx = DEFAULT_ASPECTRATIO_IDX;
		private int myResolutionIdx = DEFAULT_RESOLUTION_IDX[DEFAULT_ASPECTRATIO_IDX];
		private int myMultiplierIdx = DEFAULT_MULTIPLIER_IDX;
		private int myKernelMode = DEFAULT_KERNEL;
		private int myMsaaLevel = 8;
		private int myFileFormat = 0;

		private FreeControllerV3 mySelectedController = null;
		private float myWaitClock = 0;
		private bool myNeedSetup = true;		
		
		private SuperResolution mySuperRes = new SuperResolution();

		public override void Init()
		{
			bool isDesktopMode = !XRSettings.isDeviceActive || string.IsNullOrEmpty(XRSettings.loadedDeviceName);
			
			Utils.OnInitUI(CreateUIElement);

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>SuperShot v12</b></size>\nTakes ultra-high resolution or super-sampled "+
				"screenshots, allowing for improved hair, shadow and texture quality when taking screenshots.</color>",
				200.0f, true
			);

			myMemoryInfo = new JSONStorableString("MemoryInfo", "");
			Utils.SetupInfoOneLine(this, myMemoryInfo, true);

			Utils.SetupInfoText(this,
				"Depending on the settings, this plugin can be <b>extremely demanding</b> in terms of graphics memory! "+
				"Reduce the resolution multiplier and/or MSAA when you notice trouble like incomplete images or low performance. "+
				"\n\nMemory is only allocated for the moment when taking a screenshot and released immediately afterwards.",
				380.0f, true
			);
			
			Utils.SetupAction(this, "TakeScreenshotWithPreview", TakeScreenshotWithPreview);			
			Utils.SetupAction(this, "TakeScreenshotPlayerCamera", TakeScreenshotPlayerCamera);
			Utils.SetupAction(this, "TakeScreenshotWindowCamera", TakeScreenshotWindowCamera);
			Utils.SetupAction(this, "TakeScreenshotCustom", TakeScreenshotCustom);
			

			Utils.SetupButton(this, "Take Screenshot (with Preview)", TakeScreenshotWithPreview, false);
			if (isDesktopMode)
				Utils.SetupButton(this, "Take Screenshot (Player View)", TakeScreenshotPlayerCamera, false);
			Utils.SetupButton(this, "Take Screenshot (Window Camera)", TakeScreenshotWindowCamera, false);
			Utils.SetupButton(this, "Take Screenshot (Custom Cameras)", TakeScreenshotCustom, false);					

			JSONStorableStringChooser aspectRatioChooser = Utils.SetupStringChooser(this, "Aspect Ratio", ASPECTRATIO_NAMES, DEFAULT_ASPECTRATIO_IDX, false);		
			JSONStorableStringChooser resolutionChooserOutput = Utils.SetupStringChooser(this, "Resolution Output", RESOLUTION_NAMES[DEFAULT_ASPECTRATIO_IDX], DEFAULT_RESOLUTION_IDX[DEFAULT_ASPECTRATIO_IDX], false);
			myPortaitMode = Utils.SetupToggle(this, "Portait Mode (Flip AspectRatio)", false, false);
			JSONStorableStringChooser resolutionChooserMultiplier = Utils.SetupStringChooser(this, "Resolution Multiplier", MULTIPLIER_NAMES, DEFAULT_MULTIPLIER_IDX, false);
			
			aspectRatioChooser.setCallbackFunction += (string v) => {
				int aspectRatioIdx = ASPECTRATIO_NAMES.FindIndex((string entry) => { return entry == v; });
				if (aspectRatioIdx < 0)
					aspectRatioIdx = DEFAULT_ASPECTRATIO_IDX;
				if (aspectRatioIdx != myAspectRatioIdx)
				{
					myAspectRatioIdx = aspectRatioIdx;
					resolutionChooserOutput.choices = RESOLUTION_NAMES[myAspectRatioIdx];
					resolutionChooserOutput.valNoCallback = RESOLUTION_NAMES[myAspectRatioIdx][DEFAULT_RESOLUTION_IDX[myAspectRatioIdx]];
					resolutionChooserOutput.setCallbackFunction(resolutionChooserOutput.val);
				}
				myNeedSetup = true;
			};
			resolutionChooserOutput.setCallbackFunction += (string v) => {
				myResolutionIdx = resolutionChooserOutput.choices.FindIndex((string entry) => { return entry == v; });
				if (myResolutionIdx < 0)
					myResolutionIdx = DEFAULT_RESOLUTION_IDX[myAspectRatioIdx];
				int maxMultiplierIdx = GetMaxMultiplierIdx();
				myMultiplierIdx = Mathf.Min(maxMultiplierIdx, myMultiplierIdx);
				List<string> multiplierList = new List<string>();
				for (int i=0; i<=maxMultiplierIdx; ++i)
					multiplierList.Add(MULTIPLIER_NAMES[i]);
				resolutionChooserMultiplier.choices = multiplierList;
				resolutionChooserMultiplier.val = MULTIPLIER_NAMES[myMultiplierIdx];
				myNeedSetup = true;
			};
			myPortaitMode.setCallbackFunction += (bool v) => {
				myNeedSetup = true;
			};
			resolutionChooserMultiplier.setCallbackFunction += (string v) => {
				myMultiplierIdx = resolutionChooserMultiplier.choices.FindIndex((string entry) => { return entry == v; });
				if (myMultiplierIdx < 0)
					myMultiplierIdx = DEFAULT_MULTIPLIER_IDX;
				myMultiplierIdx = Mathf.Min(GetMaxMultiplierIdx(), myMultiplierIdx);
				myNeedSetup = true;
			};

			JSONStorableStringChooser kernelChooser = Utils.SetupStringChooser(this, "Downscale Method", KERNEL_NAMES, DEFAULT_KERNEL, false);
			kernelChooser.setCallbackFunction += (string v) => {
				myKernelMode = KERNEL_NAMES.FindIndex((string entry) => { return entry == v; });
				if (myKernelMode < 0 || myKernelMode >= KERNEL_NAMES.Count)
					myKernelMode = DEFAULT_KERNEL;
				myNeedSetup = true;
			};

			JSONStorableStringChooser msaaChooser = Utils.SetupStringChooser(this, "MSAA Override", MSAA_NAMES, DEFAULT_MSAA_IDX, false);
			msaaChooser.setCallbackFunction += (string v) => {
				int idx = MSAA_NAMES.FindIndex((string entry) => { return entry == v; });
				if (idx < 0 || idx >= MSAA_VALUES.Count)
					idx = DEFAULT_MSAA_IDX;
				myMsaaLevel = MSAA_VALUES[idx];
				myNeedSetup = true;
			};

			JSONStorableStringChooser formatChooser = Utils.SetupStringChooser(this, "Image Format", FORMAT_NAMES, DEFAULT_FORMAT, false);
			formatChooser.setCallbackFunction += (string v) => {
				myFileFormat = FORMAT_NAMES.FindIndex((string entry) => { return entry == v; });
				if (myFileFormat < 0 || myFileFormat >= FORMAT_NAMES.Count)
					myFileFormat = DEFAULT_FORMAT;
			};
			
			Utils.SetupInfoOneLine(this, "<b>General Options</b>", true);
			myPreviewScale = Utils.SetupSliderFloat(this, "Preview UI scale", 2.0f, 1.0f, 2.5f, true);
			myPreviewScale.setCallbackFunction += (float v) => {
				if (this.enabled)
					SetupScreenshotToolUI();
			};
			myLockZRotation = Utils.SetupToggle(this, "Lock Z Rotation", false, true);
			myFlashLight = Utils.SetupToggle(this, "Flashlight", true, true);
			
			
			myNeedSetup = true;
			mySuperRes.OnInit(this);
		}


		private void OnDestroy()
		{
			mySuperRes.OnDestroy();
			Utils.OnDestroyUI();
		}

		private void OnEnable()
		{
			SuperController sc = SuperController.singleton;
			myScreenshotCamera = sc.hiResScreenshotCamera;
			myPreviewCamera = Instantiate<Camera>(myScreenshotCamera, myScreenshotCamera.transform);
			myPreviewCamera.transform.localPosition = Vector3.zero;
			myPreviewCamera.transform.localRotation = Quaternion.identity;
			myPreviewCamera.targetTexture = null;
			myPreviewCamera.enabled = false;

			myNeedSetup = true;

			myOriginalScreenshotTexture = myScreenshotCamera.targetTexture;
			myScreenshotCamera.targetTexture = null; // prevent regular VaM screenshot
			myScreenshotCamera.enabled = false; // prevent expensive rendering in the background

			myScreenshotHook = myScreenshotCamera.gameObject.AddComponent<ScreenshotCameraHook>();
			myScreenshotHook.Init(this);

			myPreviewHook = myPreviewCamera.gameObject.AddComponent<PreviewCameraHook>();
			myPreviewHook.Init(this);
		}

		private void OnDisable()
		{
			mySuperRes.OnDisable();

			myScreenshotCamera.targetTexture = myOriginalScreenshotTexture;
			myScreenshotCamera.enabled = SuperController.singleton.hiResScreenshotPreview.gameObject.activeSelf;
			myScreenshotCamera = null;

			if (myScreenshotHook != null)
				Destroy(myScreenshotHook);
			if (myPreviewHook != null)
				Destroy(myPreviewHook);
			myScreenshotHook = null;
			myPreviewHook = null;

			Destroy(myPreviewCamera);
			myPreviewCamera = null;

			ResetScreenshotToolUI();
		}

		private void Setup()
		{
			Vector2Int resolution = RESOLUTION_VALUES[myAspectRatioIdx][myResolutionIdx];
			mySuperRes.myOutputWidth = myPortaitMode.val ? resolution.y : resolution.x;
			mySuperRes.myOutputHeight = myPortaitMode.val ? resolution.x : resolution.y;
			mySuperRes.myMultiplier = MULTIPLIER_VALUES[myMultiplierIdx];
			mySuperRes.myKernelMode = myKernelMode;
			mySuperRes.myMsaaLevel = myMsaaLevel;
			mySuperRes.NeedSetup();

			float memoryEstimate = mySuperRes.EstimatedMemoryUseInGB();
			if (myMemoryEstimate != memoryEstimate)
			{
				myMemoryEstimate = memoryEstimate;
				myMemoryInfo.val = "<color=#FF0000><b>Plugin VRAM usage: ~" + myMemoryEstimate + " GB.</b></color>";
			}

			myNeedSetup = false;
		}

		private void Update()
		{
			if (myNeedSetup)
				Setup();

			switch (myMode)
			{
				case MODE_BUILDIN:
					UpdateScreenshotBuildInTool();
					break;
				case MODE_PLAYER:
					UpdateScreenshotPlayerCamera();
					break;
				default:
					if (SuperController.singleton.hiResScreenshotPreview.gameObject.activeSelf)
						TakeScreenshotWithPreview(); // Screenshot Tool opened via VaM menu
					break;
			}
		}

		private void TakeScreenshotWithPreview()
		{
			if (!CanTakeScreenshot())
				return;
			SuperController.singleton.SelectModeScreenshot();
			myOverrideFoV = -1.0f;
			myOverrideTransform = null;
			myOverrideFilename = null;
			myMode = MODE_BUILDIN;
			myState = STATE_SETUP;
		}

		private void TakeScreenshotPlayerCamera()
		{
			if (!CanTakeScreenshot())
				return;
			myOverrideFoV = -1.0f;
			myOverrideTransform = null;
			myOverrideFilename = null;
			myMode = MODE_PLAYER;
			myState = STATE_SETUP;
		}
		
		private void TakeScreenshotWindowCamera()
		{
			if (!CanTakeScreenshot())
				return;			
			myOverrideFoV = -1.0f;
			myOverrideTransform = null;
			myOverrideFilename = null;
			List<Atom> atoms = GetSceneAtoms();
			Atom windowCamera = atoms.Find(a => a.type == "WindowCamera");
			CameraControl cameraControl = windowCamera?.GetStorableByID("CameraControl") as CameraControl;
			if (cameraControl != null)
			{
				myOverrideTransform = cameraControl.cameraToControl.transform;			
				myOverrideFoV = cameraControl.cameraToControl.fieldOfView;
			}
						
			myMode = MODE_PLAYER;
			myState = STATE_SETUP;
		}
		
		private void TakeScreenshotCustom()
		{
			if (!CanTakeScreenshot())
				return;
			StartCoroutine(TakeScreenshotCustomCoroutine());
		}
		
		private IEnumerator TakeScreenshotCustomCoroutine()
		{
			List<Atom> atoms = GetSceneAtoms();			
			Atom windowCamera = atoms.Find(a => a.type == "WindowCamera");
			CameraControl cameraControl = windowCamera?.GetStorableByID("CameraControl") as CameraControl;
			myOverrideFoV = cameraControl != null ? cameraControl.cameraToControl.fieldOfView : -1.0f;			
			myOverrideTransform = null;
			
			atoms.RemoveAll(a => !a.name.StartsWith("SuperShot#") || !a.on);
			atoms.Sort((Atom a, Atom b) => a.name.CompareTo(b.name));
			
			if (atoms.Count == 0)
			{
				SuperController.LogMessage(
					"\nThe SuperShot plugin can take multiple screenshots at once. For this add one or more 'Empty' atoms into your "+
					"scene and rename them to 'SuperShot#X', 'SuperShot#Y', and so on. These are used as camera positions, 'X' and 'Y' "+
					"can be anything and will appear in the filename later. Atoms can be turned on/off to enabled/disable positions. "+
					"You can use the WindowCamera preview for easier positioning. SuperShot uses the FoV of the WindowCamera for these "+
					"custom screenshots.\n");
				yield break;
			}
			
			bool wasFlashing = myFlashLight.val;
			for (int i=0; i<atoms.Count; ++i)
			{
				Atom atom = atoms[i];
				int hashIdx = atom.name.IndexOf('#');
				if (hashIdx >= 0)
					myOverrideFilename = mySuperRes.GetFilename() + atom.name.Substring(hashIdx);
				else
					myOverrideFilename = null;
				myOverrideTransform = atom.mainController.transform;
				myFlashLight.valNoCallback = (i == atoms.Count-1) ? wasFlashing : false;
				
				myMode = MODE_PLAYER;
				myState = STATE_SETUP;
				while (myMode != MODE_NONE)
					yield return null;
			}
						
			myOverrideFilename = null;
		}
		
		private bool CanTakeScreenshot()
		{
			if (!enabled || !gameObject.activeInHierarchy)
			{
				SuperController.LogMessage("\nSuperShot can only take screenshots when both the plugin and the atom are enabled.");
				return false;
			}
			else if (myMode != MODE_NONE)
			{
				return false;
			}
			
			return true;
		}

		private void UpdateScreenshotBuildInTool()
		{
			switch (myState)
			{
				case STATE_SETUP:
				{
					// enable preview camera
					const int defaultSize = 1080;
					float aspectRatio = mySuperRes.myOutputWidth / (float)mySuperRes.myOutputHeight;
					int previewWidth = aspectRatio > 1 ? Mathf.RoundToInt(defaultSize*aspectRatio) : defaultSize;
					int previewHeight = aspectRatio <= 1 ? Mathf.RoundToInt(defaultSize/aspectRatio) : defaultSize;
					myPreviewCamera.targetTexture = mySuperRes.GetPreviewTexture(previewWidth, previewHeight);
					myPreviewCamera.fieldOfView = myScreenshotCamera.fieldOfView;
					myPreviewCamera.enabled = true;
					myScreenshotCamera.enabled = false;
					SetupScreenshotToolUI();

					// compatibility with PostMagic
					PostProcessingBehaviour screenshotBehaviour = myScreenshotCamera.GetComponent<PostProcessingBehaviour>();
					PostProcessingBehaviour previewBehaviour = myPreviewCamera.GetComponent<PostProcessingBehaviour>();
					if (screenshotBehaviour == null)
					{
						if (previewBehaviour != null)
							previewBehaviour.enabled = false;
					}
					else
					{
						if (previewBehaviour == null)
							previewBehaviour = myPreviewCamera.gameObject.AddComponent<PostProcessingBehaviour>();
						previewBehaviour.enabled = screenshotBehaviour.enabled;
						previewBehaviour.profile = screenshotBehaviour.profile;
					}

					myWaitClock = 0.0f;
					myState = STATE_WAIT;
					UpdateScreenshotBuildInTool();
					return;
				}
				case STATE_WAIT:
				{
					// screenshot tool closed?
					SuperController sc = SuperController.singleton;
					if (!sc.hiResScreenshotPreview.gameObject.activeSelf)
					{
						myPreviewCamera.enabled = false;
						mySuperRes.ReleaseRenderTexture(myPreviewCamera.targetTexture);
						myPreviewCamera.targetTexture = null;
						myState = STATE_SETUP;
						myMode = MODE_NONE;
						return;
					}

					// update preview
					myPreviewCamera.fieldOfView = myScreenshotCamera.fieldOfView;

					// prevent accidental screenshot when opening the tool
					myWaitClock += Time.deltaTime;
					if (myWaitClock < 0.2f)
						return;

					// wait for user input
					if (!sc.GetRightSelect() && !sc.GetLeftSelect() && !sc.GetMouseSelect())
						return;

					// init screenshot sequence
					mySuperRes.PreRender();
					myScreenshotCamera.targetTexture = mySuperRes.GetCaptureTexture();
					myScreenshotCamera.enabled = true;
					myState = STATE_PROCESS;
					return;
				}
				case STATE_PROCESS:
				{
					// process image
					mySuperRes.ProcessImage();
					mySuperRes.SaveImage(myFileFormat, myOverrideFilename);
					mySuperRes.PostRender();
					
					// cleanup
					myScreenshotCamera.enabled = false;
					myScreenshotCamera.targetTexture = null;
					myWaitClock = 0.0f;
					myState = STATE_WAIT;
					
					// flash light
					SkyshopLightController sslc = SkyshopLightController.singleton;
					if (myFlashLight.val && sslc != null)
						sslc.Flash();
					return;
				}
			}
		}

		private void UpdateScreenshotPlayerCamera()
		{
			SuperController sc = SuperController.singleton;
			
			switch (myState)
			{
				case STATE_SETUP:
				{				
					// clear selection to avoid UI elements in the camera image.
					mySelectedController = sc.GetSelectedController();
					sc.ClearSelection();
					
					// init screenshot sequence					
					mySuperRes.PreRender();
					myScreenshotCamera.targetTexture = mySuperRes.GetCaptureTexture();
					myScreenshotCamera.fieldOfView = myOverrideFoV > 0.0f ? myOverrideFoV : sc.monitorCameraFOV;
					myScreenshotCamera.enabled = true;
					myState = STATE_PROCESS;
					return;
				}
				case STATE_WAIT:
				{
					SuperController.LogError("SuperResolution: Something is wrong!");
					myState = STATE_SETUP;
					return;
				}
				case STATE_PROCESS:
				{
					// process image
					mySuperRes.ProcessImage();
					mySuperRes.SaveImage(myFileFormat, myOverrideFilename);
					mySuperRes.PostRender();
					
					// restore selection
					if (mySelectedController != null)
						sc.SelectController(mySelectedController);
					mySelectedController = null;
					
					// flash light
					SkyshopLightController sslc = SkyshopLightController.singleton;
					if (myFlashLight.val && sslc != null)
						sslc.Flash();

					// cleanup
					myScreenshotCamera.enabled = false;
					myScreenshotCamera.targetTexture = null;
					myScreenshotCamera.fieldOfView = sc.hiResScreenShotCameraFOV;
					myState = STATE_SETUP;
					myMode = MODE_NONE;
					return;
				}
			}
		}

		private void SetupScreenshotToolUI()
		{
			Transform previewTransform = SuperController.singleton.hiResScreenshotPreview.transform;
			bool isDesktopMode = !XRSettings.isDeviceActive || string.IsNullOrEmpty(XRSettings.loadedDeviceName);

			float aspectRatio = mySuperRes.myOutputWidth / (float)mySuperRes.myOutputHeight;
			const float defaultRatio = 16.0f/9.0f;
			Vector2 scale = new Vector2(aspectRatio/defaultRatio, 1.0f) * myPreviewScale.val;
			Transform renderPlane = previewTransform.Find("Render");
			renderPlane.localScale = new Vector3(scale.x, 1.0f, scale.y);
			renderPlane.GetComponent<Renderer>().sharedMaterial.mainTexture = myPreviewCamera.targetTexture;

			RectTransform backCanvas = previewTransform.Find("BackCanvas") as RectTransform;
			const float p = 10.0f;  // padding
			const float h = 150.0f; // base height
			backCanvas.offsetMax = new Vector2(defaultRatio*h*scale.x + p, h*scale.y + p);
			backCanvas.offsetMin = -backCanvas.offsetMax;

			RectTransform panel = previewTransform.Find("ScreenShotCanvas/Panel") as RectTransform;
			panel.anchorMax = new Vector2(0.5f, 0.5f);
			panel.anchorMin = new Vector2(0.5f, 0.5f);
			panel.offsetMax = new Vector2(400, backCanvas.offsetMax.y);
			panel.offsetMin = new Vector2(-400, backCanvas.offsetMin.y);

	 		RectTransform helpText = panel.Find("HelpText") as RectTransform;
			helpText.anchorMax = new Vector2(1, 0);
			helpText.anchorMin = new Vector2(0, 0);
			helpText.offsetMax = new Vector2(0, -10);
			helpText.offsetMin = new Vector2(0, -160);
			Text helpTextLabel = helpText.GetComponent<Text>();
			helpTextLabel.alignment = TextAnchor.UpperCenter;
			helpTextLabel.fontSize = isDesktopMode ? 28 : 40;

			RectTransform fovSlider = panel.Find("HiResScrFOV") as RectTransform;
			fovSlider.anchorMax = new Vector2(0.5f, 1.0f);
			fovSlider.anchorMin = new Vector2(0.5f, 1.0f);
			fovSlider.offsetMax = new Vector2(200, 120);
			fovSlider.offsetMin = new Vector2(-200, 20);
		}

		private void ResetScreenshotToolUI()
		{
			Transform previewTransform = SuperController.singleton.hiResScreenshotPreview.transform;
			Transform renderPlane = previewTransform.Find("Render");
			renderPlane.localScale = Vector3.one;
			renderPlane.GetComponent<Renderer>().sharedMaterial.mainTexture = myOriginalScreenshotTexture;

			RectTransform backCanvas = previewTransform.Find("BackCanvas") as RectTransform;
			backCanvas.offsetMax = new Vector2(275, 160);
			backCanvas.offsetMin = -backCanvas.offsetMax;

			RectTransform panel = previewTransform.Find("ScreenShotCanvas/Panel") as RectTransform;
			panel.anchorMax = new Vector2(1.0f, 0.5f);
			panel.anchorMin = new Vector2(1.0f, 0.5f);
			panel.offsetMax = new Vector2(150, -175);
			panel.offsetMin = new Vector2(-650, -325);

	 		RectTransform helpText = panel.Find("HelpText") as RectTransform;
			helpText.anchorMax = new Vector2(1, 1);
			helpText.anchorMin = new Vector2(0, 0);
			helpText.offsetMax = new Vector2(0, 0);
			helpText.offsetMin = new Vector2(0, 0);

			Text helpTextLabel = helpText.GetComponent<Text>();
			helpTextLabel.alignment = TextAnchor.MiddleCenter;
			helpTextLabel.fontSize = 40;

			RectTransform fovSlider = panel.Find("HiResScrFOV") as RectTransform;
			fovSlider.anchorMax = new Vector2(1.0f, 1.0f);
			fovSlider.anchorMin = new Vector2(0.0f, 1.0f);
			fovSlider.offsetMax = new Vector2(-200, 460);
			fovSlider.offsetMin = new Vector2(200, 360);
		}

		private int GetMaxMultiplierIdx()
		{
			int i = MULTIPLIER_VALUES.Count-1;
			int m = MULTIPLIER_VALUES[i];
			Vector2Int resolution = RESOLUTION_VALUES[myAspectRatioIdx][myResolutionIdx];
			while (resolution.x*m > MAX_RESOLUTION || resolution.y*m > MAX_RESOLUTION)
			{
				if (i == 0)
					break;
				--i;
				m = MULTIPLIER_VALUES[i];
			}
			return i;
		}

		private class ScreenshotCameraHook : MonoBehaviour
		{
			private Camera myCamera;
			private SuperShot mySuperShot;
			private SuperResolution mySuperRes;
			private Vector3 myPosition;
			private Quaternion myRotation;

			public void Init(SuperShot plugin)
			{
				mySuperShot = plugin;
				myCamera = mySuperShot.myScreenshotCamera;
				mySuperRes = mySuperShot.mySuperRes;
			}

			private void OnPreCull()
			{
				if (mySuperShot.myOverrideTransform != null)
				{
					myPosition = myCamera.transform.position;
					myRotation = myCamera.transform.rotation;
					myCamera.transform.position = mySuperShot.myOverrideTransform.position;
					myCamera.transform.rotation = mySuperShot.myOverrideTransform.rotation;
				}
				if (mySuperShot.myLockZRotation.val)
				{
					myRotation = myCamera.transform.rotation;
					Vector3 eulerAngles = myCamera.transform.eulerAngles;
					eulerAngles.z = 0.0f;
					myCamera.transform.eulerAngles = eulerAngles;
				}
			}

			private void OnPostRender()
			{
				if (mySuperShot.myOverrideTransform != null)
				{
					myCamera.transform.position = myPosition;
					myCamera.transform.rotation = myRotation;
					mySuperShot.myOverrideTransform = null;
				}
				else if (mySuperShot.myLockZRotation.val)
				{
					myCamera.transform.rotation = myRotation;
				}
			}
		}

		private class PreviewCameraHook : MonoBehaviour
		{
			private Camera myCamera;
			private SuperShot mySuperShot;
			private Quaternion myRotation;

			public void Init(SuperShot plugin)
			{
				mySuperShot = plugin;
				myCamera = mySuperShot.myPreviewCamera;
			}

			private void OnPreCull()
			{
				if (mySuperShot.myLockZRotation.val)
				{
					myRotation = myCamera.transform.rotation;
					Vector3 eulerAngles = myCamera.transform.eulerAngles;
					eulerAngles.z = 0.0f;
					myCamera.transform.eulerAngles = eulerAngles;
				}
			}

			private void OnPostRender()
			{
				if (mySuperShot.myLockZRotation.val)
					myCamera.transform.rotation = myRotation;
			}
		}
	}
}