﻿using UnityEngine;
using UnityEngine.XR;
using UnityEngine.UI;
using UnityEngine.Events;
using UnityEngine.PostProcessing;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System;
using MeshVR;
using MK.Glow;
using static MacGruber.SuperResolution;

namespace MacGruber
{
	public class SuperVideo : MVRScript
	{
		private const int DEFAULT_MULTIPLIER_IDX = 1;
		private const int DEFAULT_KERNEL = KERNEL_LANCZOS3;
		private const int DEFAULT_MSAA_IDX = 3;

		private JSONStorableBool myCameraOn;
		private JSONStorableStringChooser myResolutionChooserMultiplier;
		private JSONStorableFloat myPreviewUIScale;
		private JSONStorableString myMemoryInfo;

		private CameraControl myCameraControl;
		private CaptureCameraHook myCaptureHook;
		private OutputCameraHook myOutputHook;
		private Camera myCaptureCamera;
		private Camera myOutputCamera;
		private Transform myPreviewTransform1;
		private Transform myPreviewTransform2;
		private Transform myHudPreviewTransform;
		private Transform myHudBackgroundTransform;
		private RenderTexture myOriginalPreviewTexture;

		private int myMultiplierIdx = DEFAULT_MULTIPLIER_IDX;
		private int myKernelMode = DEFAULT_KERNEL;
		private int myMsaaLevel = 8;
		private float myMemoryEstimate = -1.0f;
		private bool myNeedSetup = true;

		private SuperResolution mySuperRes = new SuperResolution();

		public override void Init()
		{
			myNeedSetup = true;
			
			Utils.OnInitUI(CreateUIElement);

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>SuperVideo v12</b></size>\nEnables super-sampling for the WindowCamera atom, "+
				"allowing for improved hair, shadow and texture quality when recording video. "+
				"For optimal results set your monitor to 60Hz refresh rate and use the FrameRate Control plugin to enable "+
				"VSync or VSync-Half to for a capture frame rate in OBS of 60 or 30fps.</color>",
				350.0f, true
			);

			myMemoryInfo = new JSONStorableString("MemoryInfo", "");
			Utils.SetupInfoOneLine(this, myMemoryInfo, true);

			Utils.SetupInfoText(this,
				"Depending on the settings, this plugin can be <b>extremely demanding</b> in terms of graphics memory! "+
				"Reduce the resolution multiplier and/or MSAA when you notice trouble like incomplete images or low performance.",
				200.0f, true
			);
			
			myCameraOn = Utils.SetupToggle(this, "WindowCamera On", false, false);
			myCameraOn.isStorable = myCameraOn.isRestorable = false;
			myCameraOn.setCallbackFunction += (bool v) => {
				if (myCameraControl != null)
					myCameraControl.SetBoolParamValue("cameraOn", v);
			};

			myResolutionChooserMultiplier = Utils.SetupStringChooser(this, "Resolution Multiplier", MULTIPLIER_NAMES, DEFAULT_MULTIPLIER_IDX, false);
			myResolutionChooserMultiplier.setCallbackFunction += (string v) => {
				myMultiplierIdx = myResolutionChooserMultiplier.choices.FindIndex((string entry) => { return entry == v; });
				if (myMultiplierIdx < 0)
					myMultiplierIdx = DEFAULT_MULTIPLIER_IDX;
				myMultiplierIdx = Mathf.Min(GetMaxMultiplierIdx(), myMultiplierIdx);
				myNeedSetup = true;
			};

			JSONStorableStringChooser kernelChooser = Utils.SetupStringChooser(this, "Downscale Method", KERNEL_NAMES, DEFAULT_KERNEL, false);
			kernelChooser.setCallbackFunction += (string v) => {
				myKernelMode = KERNEL_NAMES.FindIndex((string entry) => { return entry == v; });
				if (myKernelMode < 0 || myKernelMode >= KERNEL_NAMES.Count)
					myKernelMode = DEFAULT_KERNEL;
				myNeedSetup = true;
			};

			JSONStorableStringChooser msaaChooser = Utils.SetupStringChooser(this, "MSAA Override", MSAA_NAMES, DEFAULT_MSAA_IDX, false);
			msaaChooser.setCallbackFunction += (string v) => {
				int idx = MSAA_NAMES.FindIndex((string entry) => { return entry == v; });
				if (idx < 0 || idx >= MSAA_VALUES.Count)
					idx = DEFAULT_MSAA_IDX;
				myMsaaLevel = MSAA_VALUES[idx];
				myNeedSetup = true;
			};

			myPreviewUIScale = Utils.SetupSliderFloat(this, "Preview UI Scale", 1.0f, 0.5f, 2.0f, false);
			myPreviewUIScale.setCallbackFunction += (float v) => { SetupPreviewUI(); };

			mySuperRes.OnInit(this);
		}

		private void OnDestroy()
		{
			myNeedSetup = true;
			mySuperRes.OnDestroy();
			Utils.OnDestroyUI();
		}

		private void OnDisable()
		{
			myNeedSetup = true;

			ResetPreviewUI();

			if (myCameraControl != null)
			{
				myCaptureCamera.targetTexture = null;
				myCaptureCamera.enabled = true;				
				myCaptureCamera = null;
				myCameraControl.GetBoolJSONParam("cameraOn").setCallbackFunction -= OnSyncCameraOn;		
				myCameraControl = null;
			}
			if (myOutputCamera != null)
			{
				Destroy(myOutputCamera.gameObject);
				myOutputCamera = null;
				myOutputHook = null;
			}
			if (myCaptureHook != null)
			{
				Destroy(myCaptureHook);
				myCaptureHook = null;
			}
			mySuperRes.OnDisable();
		}
		
		private void OnSyncCameraOn(bool on)
		{
			myCameraOn.valNoCallback = on;
		}

		private void OnResolutionChange()
		{
			int maxMultiplierIdx = GetMaxMultiplierIdx();
			myMultiplierIdx = Mathf.Min(maxMultiplierIdx, myMultiplierIdx);
			List<string> multiplierList = new List<string>();
			for (int i=0; i<=maxMultiplierIdx; ++i)
				multiplierList.Add(MULTIPLIER_NAMES[i]);
			myResolutionChooserMultiplier.choices = multiplierList;
			myResolutionChooserMultiplier.valNoCallback = MULTIPLIER_NAMES[myMultiplierIdx];
			mySuperRes.myOutputWidth = Screen.width;
			mySuperRes.myOutputHeight = Screen.height;
			SetupPreviewUI();
			myNeedSetup = true;
		}

		private void Setup()
		{
			mySuperRes.myOutputWidth = Screen.width;
			mySuperRes.myOutputHeight = Screen.height;
			mySuperRes.myMultiplier = MULTIPLIER_VALUES[myMultiplierIdx];
			mySuperRes.myKernelMode = myKernelMode;
			mySuperRes.myMsaaLevel = myMsaaLevel;
			mySuperRes.myOutputToScreen = true;
			mySuperRes.NeedSetup();

			float memoryEstimate = mySuperRes.EstimatedMemoryUseInGB();
			if (myMemoryEstimate != memoryEstimate)
			{
				myMemoryEstimate = memoryEstimate;
				myMemoryInfo.val = "<color=#FF0000><b>Plugin VRAM usage: ~" + myMemoryEstimate + " GB.</b></color>";
			}

			if (myCameraControl == null)
			{
				List<Atom> atoms = GetSceneAtoms();
				Atom windowCamera = atoms.Find(a => a.type == "WindowCamera");
				myCameraControl = windowCamera?.GetStorableByID("CameraControl") as CameraControl;
				
				JSONStorableBool cameraOn = myCameraControl.GetBoolJSONParam("cameraOn");
				cameraOn.setCallbackFunction += OnSyncCameraOn;
				OnSyncCameraOn(cameraOn.val);
				
				myCaptureCamera = myCameraControl.cameraToControl;
				myCaptureHook = myCaptureCamera.gameObject.AddComponent<CaptureCameraHook>();
				myCaptureHook.Init(this);
				
				GameObject go = new GameObject("SuperVideo Output");
				go.transform.parent = myCaptureCamera.transform;
				go.transform.localPosition = Vector3.zero;
				go.transform.localRotation = Quaternion.identity;
				myOutputCamera = go.AddComponent<Camera>();
				myOutputCamera.CopyFrom(myCaptureCamera);
				myOutputCamera.depth = myCaptureCamera.depth + 1;
				myOutputCamera.targetTexture = null;
				myOutputCamera.clearFlags = CameraClearFlags.Nothing;
				myOutputCamera.cullingMask = 0;
				myOutputHook = go.AddComponent<OutputCameraHook>();
				myOutputHook.Init(this);
								
				myPreviewTransform1 = myCameraControl.otherObjectToControl?.transform.GetChild(0); // "Plane"
				myPreviewTransform2 = myCameraControl.otherObjectToControl?.transform.GetChild(1); // "Plane"
				myHudPreviewTransform = myCameraControl.HUDView?.transform.Find("PlaneForCamera");
				myHudBackgroundTransform = myCameraControl.HUDView?.transform.Find("Background");
				
				myOriginalPreviewTexture = myCameraControl.cameraToControl2.targetTexture;
				myCameraControl.cameraToControl2.targetTexture = null;
				
				SetupPreviewUI();
			}

			myNeedSetup = false;
		}
		
		private void SetupPreviewUI()
		{
			if (myPreviewTransform1 == null)
				return;
			
			// preview texture
			RenderTexture previewTexture = myCameraControl.cameraToControl2.targetTexture;
			if (previewTexture != null)
				mySuperRes.ReleaseRenderTexture(previewTexture);
			const int defaultSize = 720;
			float aspectRatio = mySuperRes.myOutputWidth / (float)mySuperRes.myOutputHeight;
			int previewWidth = aspectRatio > 1 ? Mathf.RoundToInt(defaultSize*aspectRatio) : defaultSize;
			int previewHeight = aspectRatio <= 1 ? Mathf.RoundToInt(defaultSize/aspectRatio) : defaultSize;
			myCameraControl.cameraToControl2.targetTexture = previewTexture = mySuperRes.GetPreviewTexture(previewWidth, previewHeight);
			myPreviewTransform1.GetComponent<Renderer>().sharedMaterial.mainTexture = previewTexture;
			myPreviewTransform2.GetComponent<Renderer>().sharedMaterial.mainTexture = previewTexture;
			myHudPreviewTransform.GetComponent<Renderer>().sharedMaterial.mainTexture = previewTexture;			

			// 3D preview plane
			float aspect = mySuperRes.myOutputWidth / (float)mySuperRes.myOutputHeight;
			float size = 0.018f * myPreviewUIScale.val;
			Vector3 scale = new Vector3(size, 1.0f, size);
			if (aspect > 1.0f)
				scale.x *= aspect;
			else
				scale.z /= aspect;
			myPreviewTransform1.localScale = scale;			
			myPreviewTransform2.localScale = scale;

			// HUD preview
			const float hudsize = 0.0125f;
			Vector3 hudscale = new Vector3(hudsize, 1.0f, hudsize);
			if (aspect > 1.0f)
				hudscale.x *= aspect;
			else
				hudscale.z /= aspect;
			Vector3 hudoffset = new Vector3(-0.2f, 0.19875f+hudscale.z*2.5f, 0.0f);
			myHudPreviewTransform.localScale = hudscale;
			myHudPreviewTransform.localPosition = hudoffset;
			myHudBackgroundTransform.localScale = hudscale + new Vector3(0.001f, 0.0f, 0.001f);
			myHudBackgroundTransform.localPosition = new Vector3(hudoffset.x, hudoffset.y, -0.002f);
		}

		private void ResetPreviewUI()
		{
			if (myPreviewTransform1 == null)
				return;
			
			// preview texture
			if (myCameraControl.cameraToControl2.targetTexture != null)
				mySuperRes.ReleaseRenderTexture(myCameraControl.cameraToControl2.targetTexture);
			myCameraControl.cameraToControl2.targetTexture = myOriginalPreviewTexture;
			myPreviewTransform1.GetComponent<Renderer>().sharedMaterial.mainTexture = myOriginalPreviewTexture;
			myPreviewTransform2.GetComponent<Renderer>().sharedMaterial.mainTexture = myOriginalPreviewTexture;
			myHudPreviewTransform.GetComponent<Renderer>().sharedMaterial.mainTexture = myOriginalPreviewTexture;

			// 3D preview plane
			const float aspect = 16.0f/9.0f;
			const float size = 0.018f;
			Vector3 scale = new Vector3(aspect*size, size, size);
			myPreviewTransform1.localScale = scale;
			myPreviewTransform2.localScale = scale;

			// HUD preview
			myHudPreviewTransform.localScale = new Vector3(0.02222f, 0.0125f, 0.0125f);
			myHudPreviewTransform.localPosition = new Vector3(-0.2f, 0.23f, 0.0f);
            myHudBackgroundTransform.localScale = new Vector3(0.02322f, 0.0135f, 0.0135f);
			myHudBackgroundTransform.localPosition = new Vector3(-0.2f, 0.23f, -0.002f);
		}

		private void Update()
		{
			bool isActive = 
				myCameraControl != null &&
				myCameraControl.cameraGroup.gameObject.activeSelf &&
				!SuperController.singleton.MonitorRig.gameObject.activeSelf;

			if (mySuperRes.myOutputWidth != Screen.width || mySuperRes.myOutputHeight != Screen.height)
				OnResolutionChange();
			if (myNeedSetup)
				Setup();
			
			// Prevent expensive rendering in the background
			bool useOutputCamera = mySuperRes.myMultiplier > 1;
			myCaptureCamera.enabled = isActive;
			myOutputCamera.enabled = isActive && useOutputCamera;
			
			// Override MSAA while active
			int msaaLevel = 0;
			if (isActive)
				msaaLevel = useOutputCamera ? 0 : myMsaaLevel;
			else
				msaaLevel = UserPreferences.singleton.msaaLevel;
			if (QualitySettings.antiAliasing != msaaLevel)
				QualitySettings.antiAliasing = msaaLevel;
		}

		private int GetMaxMultiplierIdx()
		{
			int i = MULTIPLIER_VALUES.Count-1;
			int m = MULTIPLIER_VALUES[i];
			while (Screen.width*m > MAX_RESOLUTION || Screen.height*m > MAX_RESOLUTION)
			{
				if (i == 0)
					break;
				--i;
				m = MULTIPLIER_VALUES[i];
			}
			return i;
		}
		
		private class CaptureCameraHook : MonoBehaviour
		{
			private Camera myCamera;
			private SuperResolution mySuperRes;

			public void Init(SuperVideo plugin)
			{
				myCamera = plugin.myCaptureCamera;
				mySuperRes = plugin.mySuperRes;
			}
			
			private void OnPreRender()
			{
				mySuperRes.PreRender();
				myCamera.targetTexture = mySuperRes.GetCaptureTexture();
			}
			
			private void OnPostRender()
			{
				if (mySuperRes.myMultiplier == 1)
				{
					mySuperRes.ProcessImage();	
					mySuperRes.PostRender();
					myCamera.targetTexture = null;
				}
			}
		}

		private class OutputCameraHook : MonoBehaviour
		{
			private SuperResolution mySuperRes;

			public void Init(SuperVideo plugin)
			{
				mySuperRes = plugin.mySuperRes;
			}
			
			private void OnPostRender()
			{
				mySuperRes.ProcessImage();
				mySuperRes.PostRender();
			}
		}
	}
}